from functools import partial, wraps
from inspect import signature
from itertools import combinations

import numpy as np
from scipy.ndimage import label
from scipy.stats.mstats import linregress
from scipy.fftpack import ifft

from .minimum_phase_decomposition import minimum_phase_decomposition
from .statistics import (adjust_for_multiple_comparisons,
                         fisher_z_transform,
                         get_normal_distribution_p_values, coherence_bias)

EXPECTATION = {
    'trials': partial(np.mean, axis=1),
    'tapers': partial(np.mean, axis=2),
    'trials_tapers': partial(np.mean, axis=(1, 2))
}


class lazyproperty:
    '''Computes property if it hasn't been computed, otherwise stores the
    answer and returns the answer.

    Useful if property computation is expensive. To use, simply use as a
    decorator as one would use `property`.

    '''
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            value = self.func(instance)
            setattr(instance, self.func.__name__, value)
            return value


def non_negative_frequencies(axis):
    '''Decorator that removes the negative frequencies.'''
    def decorator(connectivity_measure):
        @wraps(connectivity_measure)
        def wrapper(*args, **kwargs):
            measure = connectivity_measure(*args, **kwargs)
            if measure is not None:
                n_frequencies = measure.shape[axis]
                non_neg_index = np.arange(0, (n_frequencies + 1) // 2)
                return np.take(measure, indices=non_neg_index, axis=axis)
            else:
                return None
        return wrapper
        wrapper.__docstring__ = connectivity_measure.__docstring__
    return decorator


class Connectivity(object):
    '''Computes brain connectivity measures based on the cross spectral
    matrix.

    Spectral granger methods that require estimation of transfer function
    and noise covariance use minimum phase decomposition [1] to decompose
    the cross spectral matrix into square roots, which then can be used to
    non-parametrically estimate the transfer function and noise covariance.

    Attributes
    ----------
    fourier_coefficients : array, shape (n_time_windows, n_trials,
                                         n_tapers, n_fft_samples,
                                         n_signals)
        The compex-valued coefficients from a fourier transform. Note that
        this is expected to be the two-sided fourier coefficients
        (both the positive and negative lags). This is needed for the
        Granger-based methods to work.
    expectation_type : ('trials_tapers' | 'trials' | 'tapers')
        How to average the cross spectral matrix. 'trials_tapers' averages
        over the trials and tapers dimensions. 'trials' only averages over
        the trials dimensions (leaving tapers) and 'tapers' only averages
        over tapers (leaving trials).
    frequencies : array, shape (n_fft_samples,)
    time : array, shape (n_time_windows,)

    Methods
    -------
    coherency
    coherence_magnitude
    coherence_phase
    canonical_coherence
    imaginary_coherence
    phase_locking_value
    phase_lag_index
    weighted_phase_lag_index
    debiased_squared_phase_lag_index
    debiased_squared_weighted_phase_lag_index
    pairwise_phase_consistency
    directed_transfer_function
    directed_coherence
    partial_directed_coherence
    generalized_partial_directed_coherence
    direct_directed_transfer_function
    group_delay
    phase_lag_index
    pairwise_spectral_granger_prediction
    conditional_spectral_granger_prediction (Not implemented)
    blockwise_spectral_granger_prediction (Not implemented)

    References
    ----------
    .. [1] Dhamala, M., Rangarajan, G., and Ding, M. (2008). Analyzing
           information flow in brain networks with nonparametric Granger
           causality. NeuroImage 41, 354-362.

    '''

    def __init__(self, fourier_coefficients,
                 expectation_type='trials_tapers', frequencies=None,
                 time=None):
        self.fourier_coefficients = fourier_coefficients
        self.expectation_type = expectation_type
        self._frequencies = frequencies
        self.time = time

    @classmethod
    def from_multitaper(cls, multitaper_instance,
                        expectation_type='trials_tapers'):
        '''Construct connectivity class using a multitaper instance'''
        return cls(
            fourier_coefficients=multitaper_instance.fft(),
            expectation_type=expectation_type,
            time=multitaper_instance.time,
            frequencies=multitaper_instance.frequencies
        )

    @property
    @non_negative_frequencies(axis=0)
    def frequencies(self):
        if self._frequencies is not None:
            return self._frequencies

    @lazyproperty
    def _power(self):
        return self._expectation(
            self.fourier_coefficients *
            self.fourier_coefficients.conjugate()).real

    @lazyproperty
    def _cross_spectral_matrix(self):
        '''The complex-valued linear association between fourier
        coefficients at each frequency.

        Parameters
        ----------
        fourier_coefficients : array, shape (n_time_windows, n_trials,
                                             n_tapers, n_fft_samples,
                                             n_signals)

        Returns
        -------
        cross_spectral_matrix : array, shape (n_time_windows, n_trials,
                                              n_tapers, n_fft_samples,
                                              n_signals, n_signals)

        '''
        fourier_coefficients = self.fourier_coefficients[..., np.newaxis]
        return _complex_inner_product(fourier_coefficients,
                                      fourier_coefficients)

    @lazyproperty
    def _minimum_phase_factor(self):
        return minimum_phase_decomposition(
            self._expectation(self._cross_spectral_matrix))

    @lazyproperty
    @non_negative_frequencies(axis=-3)
    def _transfer_function(self):
        return _estimate_transfer_function(self._minimum_phase_factor)

    @lazyproperty
    def _noise_covariance(self):
        return _estimate_noise_covariance(self._minimum_phase_factor)

    @lazyproperty
    def _MVAR_Fourier_coefficients(self):
        return np.linalg.inv(self._transfer_function)

    @property
    def _expectation(self):
        return EXPECTATION[self.expectation_type]

    @property
    def n_observations(self):
        axes = signature(self._expectation).parameters['axis'].default
        if isinstance(axes, int):
            return self.fourier_coefficients.shape[axes]
        else:
            return np.prod(
                [self.fourier_coefficients.shape[axis]
                 for axis in axes])

    @non_negative_frequencies(axis=-2)
    def power(self):
        return self._power

    @non_negative_frequencies(axis=-3)
    def coherency(self):
        '''The complex-valued linear association between time series in the
         frequency domain.

         Returns
         -------
         complex_coherency : array, shape (..., n_fft_samples, n_signals,
                                           n_signals)

         '''
        norm = np.sqrt(self._power[..., :, np.newaxis] *
                       self._power[..., np.newaxis, :])
        norm[norm == 0] = np.nan
        complex_coherencey = (
            self._expectation(self._cross_spectral_matrix) / norm)
        n_signals = self.fourier_coefficients.shape[-1]
        diagonal_ind = np.arange(0, n_signals)
        complex_coherencey[..., diagonal_ind, diagonal_ind] = np.nan
        return complex_coherencey

    def coherence_phase(self):
        '''The phase angle of the complex coherency.

        Returns
        -------
        phase : array, shape (..., n_fft_samples, n_signals, n_signals)

        '''
        return np.angle(self.coherency())

    def coherence_magnitude(self):
        '''The magnitude of the complex coherency.

        Note that this is not the magnitude squared coherence.

        Returns
        -------
        magnitude : array, shape (..., n_fft_samples, n_signals, n_signals)

        '''
        return _squared_magnitude(self.coherency())

    @non_negative_frequencies(axis=-3)
    def imaginary_coherence(self):
        '''The normalized imaginary component of the cross-spectrum.

        Projects the cross-spectrum onto the imaginary axis to mitigate the
        effect of volume-conducted dependencies. Assumes volume-conducted
        sources arrive at sensors at the same time, resulting in
        a cross-spectrum with phase angle of 0 (perfectly in-phase) or \pi
        (anti-phase) if the sensors are on opposite sides of a dipole
        source. With the imaginary coherence, in-phase and anti-phase
        associations are set to zero.

        Returns
        -------
        imaginary_coherence_magnitude : array, shape (..., n_fft_samples,
                                                      n_signals, n_signals)

        References
        ----------
        .. [1] Nolte, G., Bai, O., Wheaton, L., Mari, Z., Vorbach, S., and
               Hallett, M. (2004). Identifying true brain interaction from
               EEG data using the imaginary part of coherency. Clinical
               Neurophysiology 115, 2292-2307.

        '''
        return np.abs(
            self._expectation(self._cross_spectral_matrix).imag /
            np.sqrt(self._power[..., :, np.newaxis] *
                    self._power[..., np.newaxis, :]))

    def canonical_coherence(self, group_labels):
        '''Finds the maximal coherence between each combination of groups.

        The canonical coherence finds two sets of weights such that the
        coherence between the linear combination of group1 and the linear
        combination of group2 is maximized.

        Parameters
        ----------
        group_labels : array-like, shape (n_signals,)
            Links each signal to a group.

        Returns
        -------
        canonical_coherence : array, shape (n_time_samples, n_fft_samples,
                                            n_groups, n_groups)
            The maximimal coherence for each group pair
        labels : array, shape (n_groups,)
            The sorted unique group labels that correspond to `n_groups`

        References
        ----------
        .. [1] Stephen, E.P. (2015). Characterizing dynamically evolving
               functional networks in humans with application to speech.
               Boston University.

        '''
        labels = np.unique(group_labels)
        n_frequencies = self.fourier_coefficients.shape[-2]
        non_negative_frequencies = np.arange(0, (n_frequencies + 1) // 2)
        fourier_coefficients = self.fourier_coefficients[
            ..., non_negative_frequencies, :]
        normalized_fourier_coefficients = [
            _normalize_fourier_coefficients(
                fourier_coefficients[..., np.in1d(group_labels, label)])
            for label in labels]

        n_groups = len(labels)
        new_shape = (self.time.size, self.frequencies.size, n_groups,
                     n_groups)
        magnitude = _squared_magnitude(np.stack([
            _estimate_canonical_coherence(
                fourier_coefficients1, fourier_coefficients2)
            for fourier_coefficients1, fourier_coefficients2
            in combinations(normalized_fourier_coefficients, 2)
        ], axis=-1))

        canonical_coherence_magnitude = np.full(new_shape, np.nan)
        group_combination_ind = np.array(
            list(combinations(np.arange(n_groups), 2)))
        canonical_coherence_magnitude[
            ..., group_combination_ind[:, 0],
            group_combination_ind[:, 1]] = magnitude
        canonical_coherence_magnitude[
            ..., group_combination_ind[:, 1],
            group_combination_ind[:, 0]] = magnitude

        return canonical_coherence_magnitude, labels

    @non_negative_frequencies(axis=-3)
    def phase_locking_value(self):
        '''The cross-spectrum with the power for each signal scaled to
        a magnitude of 1.

        The phase locking value attempts to mitigate power differences
        between realizations (tapers or trials) by treating all values of
        the cross-spectrum as the same power. This has the effect of
        downweighting high power realizations and upweighting low power
        realizations.

        Returns
        -------
        phase_locking_value : array, shape (..., n_fft_samples, n_signals,
                                            n_signals)

        References
        ----------
        .. [1] Lachaux, J.-P., Rodriguez, E., Martinerie, J., Varela, F.J.,
               and others (1999). Measuring phase synchrony in brain
               signals. Human Brain Mapping 8, 194-208.

        '''
        return self._expectation(
            self._cross_spectral_matrix /
            np.abs(self._cross_spectral_matrix))

    @non_negative_frequencies(axis=-3)
    def phase_lag_index(self):
        '''A non-parametric synchrony measure designed to mitigate power
        differences between realizations (tapers, trials) and
        volume-conduction.

        The phase lag index is the average sign of the imaginary
        component of the cross-spectrum. The imaginary component sets
        in-phase or anti-phase signals to zero and the sign scales it to
        have the same magnitude regardless of phase.

        Returns
        -------
        phase_lag_index : array, shape (..., n_fft_samples, n_signals,
                                        n_signals)

        References
        ----------
        .. [1] Stam, C.J., Nolte, G., and Daffertshofer, A. (2007). Phase
               lag index: Assessment of functional connectivity from multi
               channel EEG and MEG with diminished bias from common
               sources. Human Brain Mapping 28, 1178-1193.

        '''
        return self._expectation(np.sign(self._cross_spectral_matrix.imag))

    @non_negative_frequencies(-3)
    def weighted_phase_lag_index(self):
        '''Weighted average of the phase lag index using the imaginary
        coherency magnitudes as weights.

        Returns
        -------
        weighted_phase_lag_index : array, shape (..., n_fft_samples,
                                                 n_signals, n_signals)

        References
        ----------
        .. [1] Vinck, M., Oostenveld, R., van Wingerden, M., Battaglia, F.,
               and Pennartz, C.M.A. (2011). An improved index of
               phase-synchronization for electrophysiological data in the
               presence of volume-conduction, noise and sample-size bias.
               NeuroImage 55, 1548-1565.

        '''
        weights = self._expectation(
            np.abs(self._cross_spectral_matrix.imag))
        weights[weights < np.finfo(float).eps] = 1
        return self._expectation(
            self._cross_spectral_matrix.imag) / weights

    def debiased_squared_phase_lag_index(self):
        '''The square of the phase lag index corrected for the positive
        bias induced by using the magnitude of the complex cross-spectrum.

        Returns
        -------
        phase_lag_index : array, shape (..., n_fft_samples, n_signals,
                                        n_signals)

        References
        ----------
        .. [1] Vinck, M., Oostenveld, R., van Wingerden, M., Battaglia, F.,
               and Pennartz, C.M.A. (2011). An improved index of
               phase-synchronization for electrophysiological data in the
               presence of volume-conduction, noise and sample-size bias.
               NeuroImage 55, 1548-1565.

        '''
        n_observations = self.n_observations
        return ((n_observations * self.phase_lag_index() ** 2 - 1.0) /
                (n_observations - 1.0))

    @non_negative_frequencies(-3)
    def debiased_squared_weighted_phase_lag_index(self):
        '''The square of the weighted phase lag index corrected for the
        positive bias induced by using the magnitude of the complex
        cross-spectrum.

        Returns
        -------
        weighted_phase_lag_index : array, shape (..., n_fft_samples,
                                                 n_signals, n_signals)

        References
        ----------
        .. [1] Vinck, M., Oostenveld, R., van Wingerden, M., Battaglia, F.,
               and Pennartz, C.M.A. (2011). An improved index of
               phase-synchronization for electrophysiological data in the
               presence of volume-conduction, noise and sample-size bias.
               NeuroImage 55, 1548-1565.

        '''
        n_observations = self.n_observations
        imaginary_cross_spectral_matrix_sum = self._expectation(
            self._cross_spectral_matrix.imag) * n_observations
        squared_imaginary_cross_spectral_matrix_sum = self._expectation(
            self._cross_spectral_matrix.imag ** 2) * n_observations
        imaginary_cross_spectral_matrix_magnitude_sum = self._expectation(
            np.abs(self._cross_spectral_matrix.imag)) * n_observations
        weights = (imaginary_cross_spectral_matrix_magnitude_sum ** 2 -
                   squared_imaginary_cross_spectral_matrix_sum)
        weights[weights == 0] = np.nan
        return (imaginary_cross_spectral_matrix_sum ** 2 -
                squared_imaginary_cross_spectral_matrix_sum) / weights

    def pairwise_phase_consistency(self):
        '''The square of the phase locking value corrected for the
        positive bias induced by using the magnitude of the complex
        cross-spectrum.

        Returns
        -------
        phase_locking_value : array, shape (..., n_fft_samples, n_signals,
                                            n_signals)

        References
        ----------
        .. [1] Vinck, M., van Wingerden, M., Womelsdorf, T., Fries, P., and
               Pennartz, C.M.A. (2010). The pairwise phase consistency: A
               bias-free measure of rhythmic neuronal synchronization.
               NeuroImage 51, 112-122.

        '''
        n_observations = self.n_observations
        plv_sum = self.phase_locking_value() * n_observations
        ppc = ((plv_sum * plv_sum.conjugate() - n_observations) /
               (n_observations ** 2 - n_observations))
        return ppc.real

    def pairwise_spectral_granger_prediction(self):
        '''The amount of power at a node in a frequency explained by (is
        predictive of) the power at other nodes.

        Also known as spectral granger causality.

        References
        ----------
        .. [1] Geweke, J. (1982). Measurement of Linear Dependence and
               Feedback Between Multiple Time Series. Journal of the
               American Statistical Association 77, 304.

        '''
        cross_spectral_matrix = self._expectation(
            self._cross_spectral_matrix)
        n_signals = cross_spectral_matrix.shape[-1]
        transfer_function = np.empty_like(cross_spectral_matrix)
        total_power = self.power()
        n_frequencies = cross_spectral_matrix.shape[-3]
        non_neg_index = np.arange(0, (n_frequencies + 1) // 2)
        new_shape = list(cross_spectral_matrix.shape)
        new_shape[-3] = non_neg_index.size
        predictive_power = np.empty(new_shape)

        for pair_indices in combinations(range(n_signals), 2):
            pair_indices = np.array(pair_indices)[:, np.newaxis]
            try:
                minimum_phase_factor = (
                    minimum_phase_decomposition(
                        cross_spectral_matrix[
                            ..., pair_indices, pair_indices.T]))
                transfer_function = _estimate_transfer_function(
                    minimum_phase_factor)[..., non_neg_index, :, :]
                rotated_covariance = _remove_instantaneous_causality(
                    _estimate_noise_covariance(minimum_phase_factor))
                predictive_power[..., pair_indices, pair_indices.T] = (
                    _estimate_predictive_power(
                        total_power[..., pair_indices[:, 0]],
                        rotated_covariance, transfer_function))
            except np.linalg.LinAlgError:
                predictive_power[
                    ..., pair_indices, pair_indices.T] = np.nan

        diagonal_ind = np.diag_indices(n_signals)
        predictive_power[..., diagonal_ind[0], diagonal_ind[1]] = np.nan

        return np.log(predictive_power)

    def conditional_spectral_granger_prediction():
        raise NotImplementedError

    def blockwise_spectral_granger_prediction():
        raise NotImplementedError

    def directed_transfer_function(self):
        '''The transfer function coupling strength normalized by the total
        influence of other signals on that signal (inflow).

        Characterizes the direct and indirect coupling to a node.

        Returns
        -------
        directed_transfer_function : array, shape (..., n_fft_samples,
                                                   n_signals, n_signals)

        References
        ----------
        .. [1] Kaminski, M., and Blinowska, K.J. (1991). A new method of
               the description of the information flow in the brain
               structures. Biological Cybernetics 65, 203-210.

        '''
        return _squared_magnitude(
            self._transfer_function /
            _total_inflow(self._transfer_function))

    def directed_coherence(self):
        '''The transfer function coupling strength normalized by the total
        influence of other signals on that signal (inflow).

        This measure is the same as the directed transfer function but the
        signal inflow is scaled by the noise variance.

        Returns
        -------
        directed_coherence : array, shape (..., n_fft_samples,
                                           n_signals, n_signals)

        References
        ----------
        .. [1] Baccala, L., Sameshima, K., Ballester, G., Do Valle, A., and
               Timo-Iaria, C. (1998). Studying the interaction between
               brain structures via directed coherence and Granger
               causality. Applied Signal Processing 5, 40.

        '''
        noise_variance = _get_noise_variance(self._noise_covariance)
        return (np.sqrt(noise_variance) *
                _squared_magnitude(self._transfer_function) /
                _total_inflow(self._transfer_function, noise_variance))

    def partial_directed_coherence(self):
        '''The transfer function coupling strength normalized by its
        strength of coupling to other signals (outflow).

        The partial directed coherence tries to regress out the influence
        of other observed signals, leaving only the direct coupling between
        two signals.

        Returns
        -------
        partial_directed_coherence : array, shape (..., n_fft_samples,
                                                   n_signals, n_signals)

        References
        ----------
        .. [1] Baccala, L.A., and Sameshima, K. (2001). Partial directed
               coherence: a new concept in neural structure determination.
               Biological Cybernetics 84, 463-474.

        '''
        return _squared_magnitude(
            self._MVAR_Fourier_coefficients /
            _total_outflow(self._MVAR_Fourier_coefficients))

    def generalized_partial_directed_coherence(self):
        '''The transfer function coupling strength normalized by its
        strength of coupling to other signals (outflow).

        The partial directed coherence tries to regress out the influence
        of other observed signals, leaving only the direct coupling between
        two signals.

        The generalized partial directed coherence scales the relative
        strength of coupling by the noise variance.

        Returns
        -------
        generalized_partial_directed_coherence : array,
                                                 shape (..., n_fft_samples,
                                                        n_signals,
                                                        n_signals)

        References
        ----------
        .. [1] Baccala, L.A., Sameshima, K., and Takahashi, D.Y. (2007).
               Generalized partial directed coherence. In Digital Signal
               Processing, 2007 15th International Conference on, (IEEE),
               pp. 163-166.

        '''
        noise_variance = _get_noise_variance(self._noise_covariance)
        return _squared_magnitude(
            self._MVAR_Fourier_coefficients /
            np.sqrt(noise_variance) / _total_outflow(
                    self._MVAR_Fourier_coefficients, noise_variance))

    def direct_directed_transfer_function(self):
        '''A combination of the directed transfer function estimate of
        directional influence between signals and the partial coherence's
        accounting for the influence of other signals.

        Returns
        -------
        direct_directed_transfer_function : array, shape
                                            (..., n_fft_samples,
                                             n_signals, n_signals)

        References
        ----------
        .. [1] Korzeniewska, A., Manczak, M., Kaminski,
               M., Blinowska, K.J., and Kasicki, S. (2003). Determination
               of information flow direction among brain structures by a
               modified directed transfer function (dDTF) method.
               Journal of Neuroscience Methods 125, 195-207.

        '''
        full_frequency_DTF = (
            self._transfer_function /
            _total_inflow(self._transfer_function, axis=(-1, -3)))
        return (np.abs(full_frequency_DTF) *
                np.sqrt(self.partial_directed_coherence()))

    def group_delay(self, frequencies_of_interest=None,
                    frequency_resolution=None,
                    significance_threshold=0.05):
        '''The average time-delay of a broadband signal.

        Parameters
        ----------
        frequencies_of_interest : array-like, shape (2,)
        frequencies : array-like, shape (n_fft_samples,)
        frequency_resolution : float

        Returns
        -------
        delay : array, shape (..., n_signals, n_signals)
        slope : array, shape (..., n_signals, n_signals)
        r_value : array, shape (..., n_signals, n_signals)

        References
        ----------
        .. [1] Gotman, J. (1983). Measurement of small time differences
               between EEG channels: method and application to epileptic
               seizure propagation. Electroencephalography and Clinical
               Neurophysiology 56, 501-514.

        '''
        frequencies = self.frequencies
        frequency_difference = frequencies[1] - frequencies[0]
        independent_frequency_step = _get_independent_frequency_step(
            frequency_difference, frequency_resolution)
        bandpassed_coherency, bandpassed_frequencies = _bandpass(
            self.coherency(), frequencies, frequencies_of_interest)
        bias = coherence_bias(self.n_observations)

        n_signals = bandpassed_coherency.shape[-1]
        signal_combination_ind = np.array(
            list(combinations(np.arange(n_signals), 2)))
        bandpassed_coherency = bandpassed_coherency[
            ..., signal_combination_ind[:, 0],
            signal_combination_ind[:, 1]]

        is_significant = _find_significant_frequencies(
            bandpassed_coherency, bias, independent_frequency_step,
            significance_threshold=significance_threshold)
        coherence_phase = np.ma.masked_array(
            np.unwrap(np.angle(bandpassed_coherency), axis=-2),
            mask=~is_significant)

        def _linear_regression(response):
            return linregress(bandpassed_frequencies, y=response)

        regression_results = np.ma.apply_along_axis(
            _linear_regression, -2, coherence_phase)
        new_shape = (
            *bandpassed_coherency.shape[:-2], n_signals, n_signals)
        slope = np.full(new_shape, np.nan)
        slope[..., signal_combination_ind[:, 0],
              signal_combination_ind[:, 1]] = np.array(
              regression_results[..., 0, :], dtype=np.float)
        slope[..., signal_combination_ind[:, 1],
              signal_combination_ind[:, 0]] = -1 * np.array(
              regression_results[..., 0, :], dtype=np.float)

        delay = slope / (2 * np.pi)

        r_value = np.ones(new_shape)
        r_value[..., signal_combination_ind[:, 0],
                signal_combination_ind[:, 1]] = np.array(
                regression_results[..., 2, :], dtype=np.float)
        r_value[..., signal_combination_ind[:, 1],
                signal_combination_ind[:, 0]] = np.array(
                regression_results[..., 2, :], dtype=np.float)
        return delay, slope, r_value

    def delay(self, frequencies_of_interest=None,
              frequency_resolution=None,
              significance_threshold=0.05, n_range=3):
        '''Find a range of possible delays from the coherence phase.

        The delay (and phase) at each frequency is indistinguishable from
        2 \pi phase jumps, but we can look at a range of possible delays
        and see which one is most likely.

        Parameters
        ----------
        frequencies_of_interest : array-like, shape (2,)
        frequencies : array-like, shape (n_fft_samples,)
        frequency_resolution : float
        n_range : int
            Number of phases to consider.

        Returns
        -------
        possible_delays : array, shape (..., n_frequencies,
                                        (n_range * 2) + 1, n_signals,
                                        n_signals)

        '''
        frequencies = self.frequencies
        frequency_difference = frequencies[1] - frequencies[0]
        independent_frequency_step = _get_independent_frequency_step(
            frequency_difference, frequency_resolution)
        bandpassed_coherency, bandpassed_frequencies = _bandpass(
            self.coherency(), frequencies, frequencies_of_interest)
        bias = coherence_bias(self.n_observations)
        n_signals = bandpassed_coherency.shape[-1]
        signal_combination_ind = np.array(
            list(combinations(np.arange(n_signals), 2)))
        bandpassed_coherency = bandpassed_coherency[
            ..., signal_combination_ind[:, 0],
            signal_combination_ind[:, 1]]

        is_significant = _find_significant_frequencies(
            bandpassed_coherency, bias, independent_frequency_step,
            significance_threshold=significance_threshold)
        coherence_phase = np.ma.masked_array(
            np.unwrap(np.angle(bandpassed_coherency), axis=-2),
            mask=~is_significant)
        possible_range = 2 * np.pi * np.arange(-n_range, n_range + 1)
        delays = np.rollaxis((
            possible_range + coherence_phase[..., np.newaxis]) /
            (2 * np.pi), -1, -2)
        new_shape = (
            *bandpassed_coherency.shape[:-1], len(possible_range),
            n_signals, n_signals)
        possible_delays = np.full(new_shape, np.nan)
        possible_delays[..., signal_combination_ind[:, 0],
                        signal_combination_ind[:, 1]] = delays
        possible_delays[..., signal_combination_ind[:, 1],
                        signal_combination_ind[:, 0]] = -delays

        return possible_delays

    def phase_slope_index(self, frequencies_of_interest=None,
                          frequency_resolution=None):
        '''The weighted average of slopes of a broadband signal projected
        onto the imaginary axis.

        The phase slope index finds the complex weighted average of the
        coherency between frequencies where the weights correspond to the
        magnitude of the coherency at that frequency. This is projected
        on to the imaginary axis to avoid volume conduction effects.

        Parameters
        ----------
        frequencies_of_interest : array-like, shape (2,)
        frequencies : array-like, shape (n_fft_samples,)
        frequency_resolution : float

        Returns
        -------
        phase_slope_index : array, shape (..., n_signals, n_signals)

        References
        ----------
        .. [1] Nolte, G., Ziehe, A., Nikulin, V.V., Schlogl, A., Kramer,
               N., Brismar, T., and Muller, K.-R. (2008). Robustly
               Estimating the Flow Direction of Information in Complex
               Physical Systems. Physical Review Letters 100.

        '''
        frequencies = self.frequencies
        bandpassed_coherency, bandpassed_frequencies = _bandpass(
            self.coherency(), frequencies, frequencies_of_interest)

        frequency_difference = frequencies[1] - frequencies[0]
        independent_frequency_step = _get_independent_frequency_step(
            frequency_difference, frequency_resolution)
        frequency_index = np.arange(0, bandpassed_frequencies.shape[0],
                                    independent_frequency_step)
        bandpassed_coherency = bandpassed_coherency[
            ..., frequency_index, :, :]

        return _inner_combination(bandpassed_coherency).imag


def _inner_combination(data, axis=-3):
    '''Takes the inner product of all possible pairs of a
    dimension without regard to order (combinations)'''
    combination_index = np.array(
        list(combinations(range(data.shape[axis]), 2)))
    combination_slice1 = np.take(data, combination_index[:, 0], axis)
    combination_slice2 = np.take(data, combination_index[:, 1], axis)
    return (combination_slice1.conjugate() * combination_slice2).sum(
        axis=axis)


def _estimate_noise_covariance(minimum_phase):
    '''Given a matrix square root of the cross spectral matrix (
    minimum phase factor), non-parametrically estimate the noise covariance
    of a multivariate autoregressive model (MVAR).

    Parameters
    ----------
    minimum_phase : array, shape (n_time_windows, n_fft_samples,
                                  n_signals, n_signals)
        The matrix square root of a cross spectral matrix.

    Returns
    -------
    noise_covariance : array, shape (n_time_windows, n_signals, n_signals)
        The noise covariance of a MVAR model.

    References
    ----------
    .. [1] Dhamala, M., Rangarajan, G., and Ding, M. (2008). Analyzing
           information flow in brain networks with nonparametric Granger
           causality. NeuroImage 41, 354-362.

    '''
    inverse_fourier_coefficients = ifft(minimum_phase, axis=-3).real
    return _complex_inner_product(
        inverse_fourier_coefficients[..., 0, :, :],
        inverse_fourier_coefficients[..., 0, :, :]).real


def _estimate_transfer_function(minimum_phase):
    '''Given a matrix square root of the cross spectral matrix (
    minimum phase factor), non-parametrically estimate the transfer
    function of a multivariate autoregressive model (MVAR).

    Parameters
    ----------
    minimum_phase : array, shape (n_time_windows, n_fft_samples,
                                  n_signals, n_signals)
        The matrix square root of a cross spectral matrix.

    Returns
    -------
    transfer_function : array, shape (n_time_windows, n_fft_samples,
                                      n_signals, n_signals)
        The transfer function of a MVAR model.

    References
    ----------
    .. [1] Dhamala, M., Rangarajan, G., and Ding, M. (2008). Analyzing
           information flow in brain networks with nonparametric Granger
           causality. NeuroImage 41, 354-362.

    '''
    inverse_fourier_coefficients = ifft(minimum_phase, axis=-3).real
    return np.matmul(
        minimum_phase,
        np.linalg.inv(inverse_fourier_coefficients[..., 0:1, :, :]))


def _estimate_predictive_power(total_power, rotated_covariance,
                               transfer_function):
    intrinsic_power = (total_power[..., np.newaxis] -
                       rotated_covariance[..., np.newaxis, :, :] *
                       _squared_magnitude(transfer_function))
    intrinsic_power[intrinsic_power == 0] = np.finfo(float).eps
    predictive_power = total_power[..., np.newaxis] / intrinsic_power
    predictive_power[predictive_power <= 0] = np.nan
    return predictive_power


def _squared_magnitude(x):
    return np.abs(x) ** 2


def _complex_inner_product(a, b):
    '''Measures the orthogonality (similarity) of complex arrays in
    the last two dimensions.'''
    return np.matmul(a, _conjugate_transpose(b))


def _remove_instantaneous_causality(noise_covariance):
    '''Rotates the noise covariance so that the effect of instantaneous
    signals (like those caused by volume conduction) are removed.

    x -> y: var(x) - (cov(x,y) ** 2 / var(y))
    y -> x: var(y) - (cov(x,y) ** 2 / var(x))

    Parameters
    ----------
    noise_covariance : array, shape (..., n_signals, n_signals)

    Returns
    -------
    rotated_noise_covariance : array, shape (..., n_signals, n_signals)
        The noise covariance without the instantaneous causality effects.

    '''
    variance = np.diagonal(noise_covariance, axis1=-1,
                           axis2=-2)[..., np.newaxis]
    return (variance.swapaxes(-1, -2) - noise_covariance ** 2 / variance)


def _set_diagonal_to_zero(x):
    '''Sets the diaginal of the last two dimensions to zero.'''
    n_signals = x.shape[-1]
    diagonal_index = np.diag_indices(n_signals)
    x[..., diagonal_index[0], diagonal_index[1]] = 0
    return x


def _total_inflow(transfer_function, noise_variance=1.0, axis=-1):
    '''Measures the effect of incoming signals onto a node via sum of
    squares.'''
    return np.sqrt(np.sum(
        noise_variance * _squared_magnitude(transfer_function),
        keepdims=True, axis=axis))


def _get_noise_variance(noise_covariance):
    '''Extracts the noise variance from the noise covariance matrix.'''
    return np.diagonal(noise_covariance, axis1=-1, axis2=-2)[
        ..., np.newaxis, :, np.newaxis]


def _total_outflow(MVAR_Fourier_coefficients, noise_variance=1.0):
    '''Measures the effect of outgoing signals on the node via
    sum of squares.'''
    return np.sqrt(np.sum(
        _squared_magnitude(MVAR_Fourier_coefficients) / noise_variance,
        keepdims=True, axis=-2))


def _reshape(fourier_coefficients):
    '''Combine trials and tapers dimensions and move the combined dimension
    to the last axis position.

    Parameters
    ----------
    fourier_coefficients : array, shape (n_time_windows, n_trials,
                                         n_tapers, n_fft_samples,
                                         n_signals)

    Returns
    -------
    fourier_coefficients : array, shape (n_time_windows, n_fft_samples,
                                         n_signals, n_trials * n_tapers)

    '''
    (n_time_windows, _, _, n_fft_samples,
     n_signals) = fourier_coefficients.shape
    new_shape = (n_time_windows, -1, n_fft_samples, n_signals)
    return np.moveaxis(fourier_coefficients.reshape(new_shape), 1, -1)


def _normalize_fourier_coefficients(fourier_coefficients):
    '''Normalizes a group of fourier coefficients by power within group

    Parameters
    ----------
    fourier_coefficients : array, shape (n_time_windows, n_trials,
                                         n_tapers, n_fft_samples,
                                         n_signals)

    Returns
    -------
    normalized_fourier_coefficients : array, shape (n_time_windows,
                                                    n_fft_samples,
                                                    n_signals,
                                                    n_trials * n_tapers)
    '''
    U, _, V_transpose = np.linalg.svd(
        _reshape(fourier_coefficients), full_matrices=False)
    return np.matmul(U, V_transpose)


def _estimate_canonical_coherence(normalized_fourier_coefficients1,
                                  normalized_fourier_coefficients2):
    '''Finds the maximum complex correlation between groups of signals
    at each time and frequency.

    Parameters
    ----------
    normalized_fourier_coefficients1 : array, shape (n_time_windows,
                                                     n_fft_samples,
                                                     n_signals,
                                                     n_trials * n_tapers)
    normalized_fourier_coefficients2 : array, shape (n_time_windows,
                                                     n_fft_samples,
                                                     n_signals,
                                                     n_trials * n_tapers)

    Returns
    -------
    canonical_coherence : array, shape (n_time_windows, n_fft_samples)

    '''
    group_cross_spectrum = _complex_inner_product(
        normalized_fourier_coefficients1, normalized_fourier_coefficients2)
    return np.linalg.svd(group_cross_spectrum,
                         full_matrices=False, compute_uv=False)[..., 0]


def _bandpass(data, frequencies, frequencies_of_interest, axis=-3):
    '''Filters the data matrix along an axis given a maximum and minimum
    frequency of interest.

    Parameters
    ----------
    data : array, shape (..., n_fft_samples, ...)
    frequencies : array, shape (n_fft_samples,)
    frequencies_of_interest : array-like, shape (2,)

    Returns
    -------
    filtered_data : array
    filtered_frequencies : array

    '''
    frequency_index = ((frequencies_of_interest[0] < frequencies) &
                       (frequencies < frequencies_of_interest[1]))
    return (np.take(data, frequency_index.nonzero()[0], axis=axis),
            frequencies[frequency_index])


def _get_independent_frequency_step(frequency_difference,
                                    frequency_resolution):
    '''Find the number of points of a frequency axis such that they
    are statistically independent given a frequency resolution.


    Parameters
    ----------
    frequency_difference : float
        The distance between two frequency points
    frequency_resolution : float
        The ability to resolve frequency points

    Returns
    -------
    frequency_step : int
        The number of points required so that two
        frequency points are statistically independent.

    '''
    return int(np.ceil(frequency_resolution / frequency_difference))


def _find_largest_significant_group(is_significant):
    '''Finds the largest cluster of significant values over frequencies.

    If frequency value is signficant and its neighbor in the next frequency
    is also a signficant value, then they are part of the same cluster.

    If there are two clusters of the same size, the first one encountered
    is the signficant cluster. All other signficant values are set to
    false.

    Parameters
    ----------
    is_significant : bool array

    Returns
    -------
    is_significant_largest : bool array

    '''
    labeled, _ = label(is_significant)
    label_groups, label_counts = np.unique(labeled, return_counts=True)

    if not np.all(label_groups == 0):
        label_counts[0] = 0
        max_group = label_groups[np.argmax(label_counts)]
        return labeled == max_group
    else:
        return np.zeros(is_significant.shape, dtype=bool)


def _get_independent_frequencies(is_significant, frequency_step):
    '''Given a `frequency_step` that determines the distance to the next
    signficant point, sets non-distinguishable points to false.

    Parameters
    ----------
    is_significant : bool array

    Returns
    -------
    is_significant_independent : bool array

    '''
    index = is_significant.nonzero()[0]
    independent_index = index[0:len(index):frequency_step]
    return np.in1d(np.arange(0, len(is_significant)), independent_index)


def _find_largest_independent_group(is_significant, frequency_step,
                                    min_group_size=3):
    '''Finds the largest signficant cluster of frequency points and
    returns the indpendent frequency points of that cluster

    Parameters
    ----------
    is_significant : bool array
    frequency_step : int
        The number of points between each independent frequency step
    min_group_size : int
        The minimum number of points for a group to be considered

    Returns
    -------
    is_significant : bool array

    '''
    is_significant = _find_largest_significant_group(is_significant)
    is_significant = _get_independent_frequencies(
        is_significant, frequency_step)
    if sum(is_significant) < min_group_size:
        is_significant[:] = False
    return is_significant


def _find_significant_frequencies(
    coherency, bias, frequency_step=1, significance_threshold=0.05,
    min_group_size=3,
        multiple_comparisons_method='Benjamini_Hochberg_procedure'):
    '''Determines the largest significant cluster along the frequency axis.

    This function uses the fisher z-transform to determine the p-values and
    adjusts for multiple comparisons using the
    `multiple_comparisons_method`. Only independent frequencies are
    returned and there must be at least `min_group_size` frequency
    points for the cluster to be returned. If there are several significant
    groups, then only the largest group is returned.

    Parameters
    ----------
    coherency : array, shape (..., n_frequencies, n_signals, n_signals)
        The complex coherency between signals.
    bias : float
        Bias from the number of indpendent estimates of the frequency
        transform.
    frequency_step : int
        The number of points between each independent frequency step
    significance_threshold : float
        The threshold for a p-value to be considered signficant.
    min_group_size : int
        The minimum number of independent frequency points for
    multiple_comparisons_method : 'Benjamini_Hochberg_procedure' |
                                  'Bonferroni_correction'
        Procedure used to correct for multiple comparisons.

    Returns
    -------
    is_significant : bool array, shape (..., n_frequencies,
                                        n_signal_combintaions)

    '''
    z_coherence = fisher_z_transform(coherency, bias)
    p_values = get_normal_distribution_p_values(z_coherence)
    is_significant = adjust_for_multiple_comparisons(
        p_values, alpha=significance_threshold)
    return np.apply_along_axis(_find_largest_independent_group, -2,
                               is_significant, frequency_step,
                               min_group_size)


def _conjugate_transpose(x):
    '''Conjugate transpose of the last two dimensions of array x'''
    return x.swapaxes(-1, -2).conjugate()



from logging import getLogger

import numpy as np
from numpy.fft import fftfreq
from scipy import interpolate
from scipy.fftpack import fft, ifft, next_fast_len
from scipy.linalg import eigvals_banded
from scipy.signal import detrend

logger = getLogger(__name__)


class Multitaper(object):
    '''Transform time-domain signal(s) to the frequency domain by using
    multiple tapering windows.

    Attributes
    ----------
    time_series : array, shape (n_time_samples, n_trials, n_signals) or
                               (n_time_samples, n_signals)
    sampling_frequency : float, optional
        Number of samples per time unit the signal(s) are recorded at.
    time_halfbandwidth_product : float, optional
    detrend_type : string, optional
    start_time : float, optional
    time_window_duration : float, optional
    time_window_step : float, optional
    tapers : array, optional
    n_tapers : int, optional
    n_time_samples_per_window : int, optional
    n_time_samples_per_step : int, optional

    '''

    def __init__(self, time_series, sampling_frequency=1000,
                 time_halfbandwidth_product=3,
                 detrend_type='constant', time_window_duration=None,
                 time_window_step=None, n_tapers=None,  tapers=None,
                 start_time=0, n_fft_samples=None,
                 n_time_samples_per_window=None,
                 n_time_samples_per_step=None, is_low_bias=True):

        self.time_series = time_series
        self.sampling_frequency = sampling_frequency
        self.time_halfbandwidth_product = time_halfbandwidth_product
        self.detrend_type = detrend_type
        self._time_window_duration = time_window_duration
        self._time_window_step = time_window_step
        self.is_low_bias = is_low_bias
        self.start_time = start_time
        self._n_fft_samples = n_fft_samples
        self._tapers = tapers
        self._n_tapers = n_tapers
        self._n_time_samples_per_window = n_time_samples_per_window
        self._n_samples_per_time_step = n_time_samples_per_step

    def __repr__(self):
        return (
            'Multitaper('
            'sampling_frequency={0.sampling_frequency!r}, '
            'time_halfbandwidth_product={0.time_halfbandwidth_product!r}, '
            'time_window_duration={0.time_window_duration!r}, '
            'time_window_step={0.time_window_step!r}, '
            'detrend_type={0.detrend_type!r}, '
            'start_time={0.start_time}, '
            'n_tapers={0.n_tapers}'
            ')'.format(self))

    @property
    def tapers(self):
        if self._tapers is None:
            self._tapers = _make_tapers(
                self.n_time_samples_per_window, self.sampling_frequency,
                self.time_halfbandwidth_product, self.n_tapers,
                is_low_bias=self.is_low_bias)
        return self._tapers

    @property
    def time_window_duration(self):
        if self._time_window_duration is None:
            self._time_window_duration = (self.n_time_samples_per_window /
                                          self.sampling_frequency)
        return self._time_window_duration

    @property
    def time_window_step(self):
        if self._time_window_step is None:
            self._time_window_step = (self.n_time_samples_per_step /
                                      self.sampling_frequency)
        return self._time_window_step

    @property
    def n_tapers(self):
        '''Number of desired tapers.

        Note that the number of tapers may be less than this number if
        the bias of the tapers is too high (eigenvalues > 0.9)

        '''
        if self._n_tapers is None:
            return int(np.floor(
                2 * self.time_halfbandwidth_product - 1))
        return self._n_tapers

    @property
    def n_time_samples_per_window(self):
        if (self._n_time_samples_per_window is None and
                self._time_window_duration is None):
            self._n_time_samples_per_window = self.time_series.shape[0]
        elif self._time_window_duration is not None:
            self._n_time_samples_per_window = int(np.round(
                self.time_window_duration * self.sampling_frequency))
        return self._n_time_samples_per_window

    @property
    def n_fft_samples(self):
        if self._n_fft_samples is None:
            self._n_fft_samples = next_fast_len(
                self.n_time_samples_per_window)
        return self._n_fft_samples

    @property
    def frequencies(self):
        return fftfreq(self.n_fft_samples, 1.0 / self.sampling_frequency)

    @property
    def n_time_samples_per_step(self):
        '''If `time_window_step` is set, then calculate the
        `n_time_samples_per_step` based on the time window duration. If
        `time_window_step` and `n_time_samples_per_step` are both not set,
        default the window step size to the same size as the window.
        '''
        if (self._n_samples_per_time_step is None and
                self._time_window_step is None):
            self._n_samples_per_time_step = self.n_time_samples_per_window
        elif self._time_window_step is not None:
            self._n_samples_per_time_step = int(
                self.time_window_step * self.sampling_frequency)
        return self._n_samples_per_time_step

    @property
    def time(self):
        original_time = (np.arange(0, self.time_series.shape[0]) /
                         self.sampling_frequency)
        window_start_time = _sliding_window(
            original_time, self.n_time_samples_per_window,
            self.n_time_samples_per_step)[:, 0]
        return self.start_time + window_start_time

    @property
    def n_signals(self):
        return (1 if len(self.time_series.shape) < 2 else
                self.time_series.shape[-1])

    @property
    def n_trials(self):
        return (1 if len(self.time_series.shape) < 3 else
                self.time_series.shape[1])

    @property
    def frequency_resolution(self):
        return (self.time_halfbandwidth_product /
                self.time_window_duration)

    @property
    def nyquist_frequency(self):
        return self.sampling_frequency / 2

    def fft(self):
        '''Compute the fast Fourier transform using the multitaper method.

        Returns
        -------
        fourier_coefficients : array, shape (n_time_windows, n_trials,
                                             n_tapers, n_fft_samples,
                                             n_signals)

        '''
        time_series = _add_axes(self.time_series)
        time_series = _sliding_window(
            time_series, window_size=self.n_time_samples_per_window,
            step_size=self.n_time_samples_per_step, axis=0)
        time_series = detrend(time_series, type=self.detrend_type)

        logger.info(self)

        return _multitaper_fft(
            self.tapers, time_series, self.n_fft_samples,
            self.sampling_frequency).swapaxes(2, -1)


def _add_axes(time_series):
    '''If no trial or signal axes included, add one in.
    '''
    n_axes = len(time_series.shape)
    if n_axes == 1:  # add trials and signals axes
        return time_series[:, np.newaxis, np.newaxis]
    elif n_axes == 2:  # add trials axis
        return time_series[:, np.newaxis, ...]
    else:
        return time_series


def _sliding_window(data, window_size, step_size=1,
                    axis=-1, is_copy=True):
    '''
    Calculate a sliding window over a signal

    Parameters
    ----------
    data : numpy array
        The array to be slided over.
    window_size : int
        Number of samples per window
    step_size : int
        Number of samples to step the window forward. Defaults to 1.
    axis : int
        The axis to slide over. Defaults to the last axis.
    is_copy : bool
        Return strided array as copy to avoid sideffects when manipulating
        the output array.

    Returns
    -------
    data : array-like
        A matrix where row in last dimension consists of one instance
        of the sliding window.

    Notes
    -----
    - Be wary of setting `copy` to `False` as undesired sideffects with the
      output values may occurr.

    Examples
    --------
    >>> a = numpy.array([1, 2, 3, 4, 5])
    >>> _sliding_window(a, window_size=3)
    array([[1, 2, 3],
           [2, 3, 4],
           [3, 4, 5]])
    >>> _sliding_window(a, window_size=3, step_size=2)
    array([[1, 2, 3],
           [3, 4, 5]])

    References
    ----------
    .. [1] https://gist.github.com/nils-werner/9d321441006b112a4b116a8387c2
    280c

    '''
    shape = list(data.shape)
    shape[axis] = np.floor(
        (data.shape[axis] / step_size) - (window_size / step_size) + 1
    ).astype(int)
    shape.append(window_size)

    strides = list(data.strides)
    strides[axis] *= step_size
    strides.append(data.strides[axis])

    strided = np.lib.stride_tricks.as_strided(
        data, shape=shape, strides=strides)

    return strided.copy() if is_copy else strided


def _multitaper_fft(tapers, time_series, n_fft_samples,
                    sampling_frequency, axis=-2):
    '''Projects the data on the tapers and returns the discrete Fourier
    transform

    Parameters
    ----------
    tapers : array_like, shape (n_time_samples_per_window, n_tapers)
    time_series : array_like, shape (n_windows, n_trials,
                                     n_time_samples_per_window)
    n_fft_samples : int
    sampling_frequency : int

    Returns
    -------
    fourier_coefficients : array_like, shape (n_windows, n_trials, n_tapers
                                              n_fft_samples, n_signals)

    '''
    projected_time_series = (time_series[..., np.newaxis] *
                             tapers[np.newaxis, np.newaxis, ...])
    return (fft(projected_time_series, n=n_fft_samples, axis=axis) /
            sampling_frequency)


def _make_tapers(n_time_samples_per_window, sampling_frequency,
                 time_halfbandwidth_product, n_tapers, is_low_bias=True):
    '''Returns the Discrete prolate spheroidal sequences (tapers) for
    multi-taper spectral analysis.

    Parameters
    ----------
    n_time_samples_per_window : int
    sampling_frequency : int
    time_halfbandwidth_product : float
    n_tapers : int
    is_low_bias : bool
        Keep only tapers with eigenvalues > 0.9

    Returns
    -------
    tapers : array_like, shape (n_time_samples_per_window, n_tapers)

    '''
    tapers, _ = dpss_windows(
        n_time_samples_per_window, time_halfbandwidth_product, n_tapers,
        is_low_bias=is_low_bias)
    return tapers.T * np.sqrt(sampling_frequency)


def tridisolve(d, e, b, overwrite_b=True):
    '''Symmetric tridiagonal system solver, from Golub and Van Loan p157.
    .. note:: Copied from NiTime.
    Parameters
    ----------
    d : ndarray
      main diagonal stored in d[:]
    e : ndarray
      superdiagonal stored in e[:-1]
    b : ndarray
      RHS vector
    Returns
    -------
    x : ndarray
      Solution to Ax = b (if overwrite_b is False). Otherwise solution is
      stored in previous RHS vector b
    '''
    N = len(b)
    # work vectors
    dw = d.copy()
    ew = e.copy()
    if overwrite_b:
        x = b
    else:
        x = b.copy()
    for k in range(1, N):
        # e^(k-1) = e(k-1) / d(k-1)
        # d(k) = d(k) - e^(k-1)e(k-1) / d(k-1)
        t = ew[k - 1]
        ew[k - 1] = t / dw[k - 1]
        dw[k] = dw[k] - t * ew[k - 1]
    for k in range(1, N):
        x[k] = x[k] - ew[k - 1] * x[k - 1]
    x[N - 1] = x[N - 1] / dw[N - 1]
    for k in range(N - 2, -1, -1):
        x[k] = x[k] / dw[k] - ew[k] * x[k + 1]

    if not overwrite_b:
        return x


def tridi_inverse_iteration(d, e, w, x0=None, rtol=1e-8):
    '''Perform an inverse iteration.
    This will find the eigenvector corresponding to the given eigenvalue
    in a symmetric tridiagonal system.
    ..note:: Copied from NiTime.
    Parameters
    ----------
    d : array
      main diagonal of the tridiagonal system
    e : array
      offdiagonal stored in e[:-1]
    w : float
      eigenvalue of the eigenvector
    x0 : array
      initial point to start the iteration
    rtol : float
      tolerance for the norm of the difference of iterates
    Returns
    -------
    e: array
      The converged eigenvector
    '''
    eig_diag = d - w
    if x0 is None:
        x0 = np.random.randn(len(d))
    x_prev = np.zeros_like(x0)
    norm_x = np.linalg.norm(x0)
    # the eigenvector is unique up to sign change, so iterate
    # until || |x^(n)| - |x^(n-1)| ||^2 < rtol
    x0 /= norm_x
    while np.linalg.norm(np.abs(x0) - np.abs(x_prev)) > rtol:
        x_prev = x0.copy()
        tridisolve(eig_diag, e, x0)
        norm_x = np.linalg.norm(x0)
        x0 /= norm_x
    return x0


def dpss_windows(n_time_samples_per_window, time_halfbandwidth_product,
                 n_tapers, is_low_bias=True, interp_from=None,
                 interp_kind='linear'):
    '''Compute Discrete Prolate Spheroidal Sequences.

    Will give of orders [0, n_tapers-1] for a given frequency-spacing
    multiple NW and sequence length `n_time_samples_per_window`.

    Copied from NiTime and MNE-Python

    Parameters
    ----------
    n_time_samples_per_window : int
        Sequence length
    time_halfbandwidth_product : float, unitless
        Standardized half bandwidth corresponding to 2 * half_bw = BW * f0
        = BW * `n_time_samples_per_window` / dt but with dt taken as 1
    n_tapers : int
        Number of DPSS windows to return
    is_low_bias : bool
        Keep only tapers with eigenvalues > 0.9
    interp_from : int (optional)
        The tapers can be calculated using interpolation from a set of
        tapers with the same NW and n_tapers, but shorter
        n_time_samples_per_window.
        This is the length of the shorter set of tapers.
    interp_kind : str (optional)
        This input variable is passed to scipy.interpolate.interp1d and
        specifies the kind of interpolation as a string ('linear',
        'nearest', 'zero', 'slinear', 'quadratic, 'cubic') or as an integer
        specifying the order of the spline interpolator to use.

    Returns
    -------
    tapers, eigenvalues : tuple
        tapers is an array, shape (n_tapers, n_time_samples_per_window)

    Notes
    -----
    Tridiagonal form of DPSS calculation from:
    Slepian, D. Prolate spheroidal wave functions, Fourier analysis, and
    uncertainty V: The discrete case. Bell System Technical Journal,
    Volume 57 (1978), 1371430

    '''
    n_tapers = int(n_tapers)
    half_bandwidth = (float(time_halfbandwidth_product) /
                      n_time_samples_per_window)
    time_index = np.arange(n_time_samples_per_window, dtype='d')

    if interp_from is not None:
        tapers = _find_tapers_from_interpolation(
            interp_from, time_halfbandwidth_product, n_tapers,
            n_time_samples_per_window, interp_kind)
    else:
        tapers = _find_tapers_from_optimization(
            n_time_samples_per_window, time_index, half_bandwidth,
            n_tapers)

    tapers = _fix_taper_sign(tapers, n_time_samples_per_window)
    eigenvalues = _get_taper_eigenvalues(
        tapers, half_bandwidth, time_index)

    return (_get_low_bias_tapers(tapers, eigenvalues)
            if is_low_bias else (tapers, eigenvalues))


def _find_tapers_from_interpolation(
    interp_from, time_halfbandwidth_product, n_tapers,
        n_time_samples_per_window, interp_kind):
    '''Create the tapers of the smaller size `interp_from` and then
    interpolate to the larger size `n_time_samples_per_window`.'''
    smaller_tapers, _ = dpss_windows(
        interp_from, time_halfbandwidth_product, n_tapers,
        is_low_bias=False)

    return [_interpolate_taper(
        taper, interp_kind, n_time_samples_per_window)
            for taper in smaller_tapers]


def _interpolate_taper(taper, interp_kind, n_time_samples_per_window):
    interpolation_function = interpolate.interp1d(
        np.arange(taper.shape[-1]), taper, kind=interp_kind)
    interpolated_taper = interpolation_function(
        np.linspace(0, taper.shape[-1] - 1, n_time_samples_per_window,
                    endpoint=False))
    return interpolated_taper / np.sqrt(np.sum(interpolated_taper ** 2))


def _find_tapers_from_optimization(n_time_samples_per_window, time_index,
                                   half_bandwidth, n_tapers):
    '''here we want to set up an optimization problem to find a sequence
    whose energy is maximally concentrated within band
    [-half_bandwidth, half_bandwidth]. Thus,
    the measure lambda(T, half_bandwidth) is the ratio between the
    energy within that band, and the total energy. This leads to the
    eigen-system (A - (l1)I)v = 0, where the eigenvector corresponding
    to the largest eigenvalue is the sequence with maximally
    concentrated energy. The collection of eigenvectors of this system
    are called Slepian sequences, or discrete prolate spheroidal
    sequences (DPSS). Only the first K, K = 2NW/dt orders of DPSS will
    exhibit good spectral concentration
    [see http://en.wikipedia.org/wiki/Spectral_concentration_problem]

    Here I set up an alternative symmetric tri-diagonal eigenvalue
    problem such that
    (B - (l2)I)v = 0, and v are our DPSS (but eigenvalues l2 != l1)
    the main diagonal = ([n_time_samples_per_window-1-2*t]/2)**2 cos(2PIW),
    t=[0,1,2,...,n_time_samples_per_window-1] and the first off-diagonal =
    t(n_time_samples_per_window-t)/2, t=[1,2,...,
    n_time_samples_per_window-1] [see Percival and Walden, 1993]'''
    diagonal = (
        ((n_time_samples_per_window - 1 - 2 * time_index) / 2.) ** 2
        * np.cos(2 * np.pi * half_bandwidth))
    off_diag = np.zeros_like(time_index)
    off_diag[:-1] = (
        time_index[1:] * (n_time_samples_per_window - time_index[1:]) / 2.)
    # put the diagonals in LAPACK 'packed' storage
    ab = np.zeros((2, n_time_samples_per_window), dtype='d')
    ab[1] = diagonal
    ab[0, 1:] = off_diag[:-1]
    # only calculate the highest n_tapers eigenvalues
    w = eigvals_banded(
        ab, select='i',
        select_range=(n_time_samples_per_window - n_tapers,
                      n_time_samples_per_window - 1))
    w = w[::-1]

    # find the corresponding eigenvectors via inverse iteration
    t = np.linspace(0, np.pi, n_time_samples_per_window)
    tapers = np.zeros((n_tapers, n_time_samples_per_window), dtype='d')
    for taper_ind in range(n_tapers):
        tapers[taper_ind, :] = tridi_inverse_iteration(
            diagonal, off_diag, w[taper_ind],
            x0=np.sin((taper_ind + 1) * t))
    return tapers


def _fix_taper_sign(tapers, n_time_samples_per_window):
    '''By convention (Percival and Walden, 1993 pg 379)
    symmetric tapers (k=0,2,4,...) should have a positive average and
    antisymmetric tapers should begin with a positive lobe.

    Parameters
    ----------
    tapers : array, shape (n_tapers, n_time_samples_per_window)
    '''

    # Fix sign of symmetric tapers
    is_not_symmetric = tapers[::2, :].sum(axis=1) < 0
    fix_sign = is_not_symmetric * -1
    fix_sign[fix_sign == 0] = 1
    tapers[::2, :] *= fix_sign[:, np.newaxis]

    # Fix sign of antisymmetric tapers.
    # rather than test the sign of one point, test the sign of the
    # linear slope up to the first (largest) peak
    largest_peak_ind = np.argmax(
        np.abs(tapers[1::2, :n_time_samples_per_window // 2]), axis=1)
    for taper_ind, peak_ind in enumerate(largest_peak_ind):
        if np.sum(tapers[2 * taper_ind + 1, :peak_ind]) < 0:
            tapers[2 * taper_ind + 1, :] *= -1
    return tapers


def _auto_correlation(data, axis=-1):
    n_time_samples_per_window = data.shape[axis]
    n_fft_samples = next_fast_len(2 * n_time_samples_per_window - 1)
    dpss_fft = fft(data, n_fft_samples, axis=axis)
    power = dpss_fft * dpss_fft.conj()
    return np.real(ifft(power, axis=axis))


def _get_low_bias_tapers(tapers, eigenvalues):
    is_low_bias = eigenvalues > 0.9
    if not np.any(is_low_bias):
        logger.warning('Could not properly use low_bias, '
                       'keeping lowest-bias taper')
        is_low_bias = [np.argmax(eigenvalues)]
    return tapers[is_low_bias, :], eigenvalues[is_low_bias]


def _get_taper_eigenvalues(tapers, half_bandwidth, time_index):
    '''Finds the eigenvalues of the original spectral concentration
    problem using the autocorr sequence technique from Percival and Walden,
    1993 pg 390

    Parameters
    ----------
    tapers : array, shape (n_tapers, n_time_samples_per_window)
    half_bandwidth : float
    time_index : array, (n_time_samples_per_window,)

    Returns
    -------
    eigenvalues : array, shape (n_tapers,)

    '''

    ideal_filter = 4 * half_bandwidth * np.sinc(
        2 * half_bandwidth * time_index)
    ideal_filter[0] = 2 * half_bandwidth
    n_time_samples_per_window = len(time_index)
    return np.dot(
        _auto_correlation(tapers)[:, :n_time_samples_per_window],
        ideal_filter)



from logging import getLogger

import numpy as np
from scipy.fftpack import fft, ifft

logger = getLogger(__name__)


def _conjugate_transpose(x):
    '''Conjugate transpose of the last two dimensions of array x'''
    return x.swapaxes(-1, -2).conjugate()


def _get_intial_conditions(cross_spectral_matrix):
    '''Returns a guess for the minimum phase factor using the Cholesky
    factorization.

    Parameters
    ----------
    cross_spectral_matrix : array, shape (n_time_samples, ...,
                                          n_fft_samples, n_signals,
                                          n_signals)

    Returns
    -------
    minimum_phase_factor : array, shape (n_time_samples, ..., 1, n_signals,
                                         n_signals)
    '''
    try:
        return np.linalg.cholesky(
            ifft(cross_spectral_matrix, axis=-3)[..., 0:1, :, :].real
        ).swapaxes(-1, -2)
    except np.linalg.linalg.LinAlgError:
        logger.warn(
            'Computing the initial conditions using the Cholesky failed. '
            'Using a random initial condition.')

        new_shape = list(cross_spectral_matrix.shape)
        N_RAND = 1000
        new_shape[-3] = N_RAND
        random_start = np.random.standard_normal(
            size=new_shape)

        random_start = np.matmul(
            random_start, _conjugate_transpose(random_start)).mean(
                axis=-3, keepdims=True)

        return np.linalg.cholesky(random_start)


def _get_causal_signal(linear_predictor):
    '''Takes half the roots on the unit circle (zero lag) and all the roots
    inside the unit circle (positive lags).

    Gives you A_(t+1)(Z) / A_(t)(Z)
    This is the plus operator in [1]

    Parameters
    ----------
    linear_predictor : array, shape (..., n_fft_samples, n_signals,
                                     n_signals)

    Returns
    -------
    causal_part_of_linear_predictor : array, shape (..., n_fft_samples,
                                                    n_signals, n_signals)

    '''
    n_signals = linear_predictor.shape[-1]
    n_fft_samples = linear_predictor.shape[-3]
    linear_predictor_coefficients = ifft(linear_predictor, axis=-3)

    # Take half of the roots on the unit circle
    linear_predictor_coefficients[..., 0, :, :] *= 0.5

    # Make the unit circle roots upper triangular
    lower_triangular_ind = np.tril_indices(n_signals, k=-1)
    linear_predictor_coefficients[
        ..., 0, lower_triangular_ind[0], lower_triangular_ind[1]] = 0

    # Take only the roots inside the unit circle (positive lags)
    linear_predictor_coefficients[..., (n_fft_samples + 1) // 2:, :, :] = 0
    return fft(linear_predictor_coefficients, axis=-3)


def _check_convergence(current, old, tolerance=1E-8):
    '''Check convergence of Wilson algorithm at each time point.

    Parameters
    ----------
    current : array, shape (n_time_points, ...)
        Current guess.
    old : array, shape (n_time_points, ...)
        Previous guess.
    tolerance : float
        Largest difference between guesses for the matrix to be judged as
        similar.

    Returns
    -------
    is_converged : array, shape (n_time_points,)
        Boolean array that indicates whether the array has converged for
        each time point.
    '''
    n_time_points = current.shape[0]
    error = np.linalg.norm(
        np.reshape(current - old, (n_time_points, -1)), ord=np.inf, axis=1)
    return error < tolerance


def _get_linear_predictor(minimum_phase_factor, cross_spectral_matrix, I):
    '''Measure how close the minimum phase factor is to the original
    cross spectral matrix.

    Parameters
    ----------
    minimum_phase_factor : array, shape (n_time_samples, ...,
                                         n_fft_samples, n_signals,
                                         n_signals)
        The current minimum phase square root guess.
    cross_spectral_matrix : array, shape (n_time_samples, ...,
                                          n_fft_samples, n_signals,
                                          n_signals)
        The matrix to be factored.
    I : array, shape (n_signals, n_signals)
        Identity matrix.

    Returns
    -------
    linear_predictor : array, shape (n_time_samples, ..., n_fft_samples,
                                     n_signals, n_signals)
        How much to adjust for the next guess for minimum phase factor.

    '''
    covariance_sandwich_estimator = np.linalg.solve(
        minimum_phase_factor, cross_spectral_matrix)
    covariance_sandwich_estimator = np.linalg.solve(
        minimum_phase_factor,
        _conjugate_transpose(covariance_sandwich_estimator))
    return covariance_sandwich_estimator + I


def minimum_phase_decomposition(cross_spectral_matrix, tolerance=1E-8,
                                max_iterations=60):
    '''Find a minimum phase matrix square root of the cross spectral
    density using the Wilson algorithm.

    Parameters
    ----------
    cross_spectral_matrix : array, shape (n_time_samples, ...,
                                          n_fft_samples, n_signals,
                                          n_signals)
    tolerance : float
        The maximum difference between guesses.
    max_iterations : int
        The maximum number of iterations for the algorithm to converge.

    Returns
    -------
    minimum_phase_factor : array, shape (n_time_samples, ...,
                                         n_fft_samples, n_signals,
                                         n_signals)
        The square root of the `cross_spectral_matrix` where all the poles
        are inside the unit circle (minimum phase).

    '''
    n_time_points = cross_spectral_matrix.shape[0]
    n_signals = cross_spectral_matrix.shape[-1]
    I = np.eye(n_signals)
    is_converged = np.zeros(n_time_points, dtype=bool)
    minimum_phase_factor = np.zeros(cross_spectral_matrix.shape)
    minimum_phase_factor[..., :, :, :] = _get_intial_conditions(
        cross_spectral_matrix)

    for iteration in range(max_iterations):
        logger.debug(
            'iteration: {0}, {1} of {2} converged'.format(
                iteration, is_converged.sum(), len(is_converged)))
        old_minimum_phase_factor = minimum_phase_factor.copy()
        linear_predictor = _get_linear_predictor(
            minimum_phase_factor, cross_spectral_matrix, I)
        minimum_phase_factor = np.matmul(
            minimum_phase_factor, _get_causal_signal(linear_predictor))

        # If already converged at a time point, don't change.
        minimum_phase_factor[is_converged, ...] = old_minimum_phase_factor[
            is_converged, ...]
        is_converged = _check_convergence(
            minimum_phase_factor, old_minimum_phase_factor, tolerance)
        if np.all(is_converged):
            return minimum_phase_factor
    else:
        logger.warning(
            'Maximum iterations reached. {} of {} converged'.format(
                is_converged.sum(), len(is_converged)))
        return minimum_phase_factor



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    name='spectral_connectivity',
    version='0.1.0.dev0',
    license='GPL-3.0',
    description=('Frequency domain functional and directed'
                 'connectivity analysis tools for electrophysiological'
                 'data'),
    author='Eric Denovellis',
    author_email='edeno@bu.edu',
    url='https://github.com/UriEdenLab/spectral_connectivity',
    packages=find_packages(),
)



import numpy as np
from scipy.stats import norm

np.seterr(invalid='ignore')


def Benjamini_Hochberg_procedure(p_values, alpha=0.05):
    '''Corrects for multiple comparisons and returns the significant
    p-values by controlling the false discovery rate at level `alpha`
    using the Benjamani-Hochberg procedure.
    Parameters
    ----------
    p_values : array_like
    alpha : float, optional
        The expected proportion of false positive tests.
    Returns
    -------
    is_significant : boolean nd-array
        A boolean array the same shape as `p_values` indicating whether the
        null hypothesis has been rejected (True) or failed to reject
        (False).
    '''
    p_values = np.array(p_values)
    threshold_line = np.linspace(0, alpha, num=p_values.size + 1,
                                 endpoint=True)[1:]
    sorted_p_values = np.sort(p_values.flatten())
    try:
        threshold_ind = np.max(
            np.where(sorted_p_values <= threshold_line)[0])
        threshold = sorted_p_values[threshold_ind]
    except ValueError:  # There are no values below threshold
        threshold = -1
    return p_values <= threshold


def Bonferroni_correction(p_values, alpha=0.05):
    p_values = np.asarray(p_values)
    return p_values <= alpha / p_values.size


MULTIPLE_COMPARISONS = dict(
    Benjamini_Hochberg_procedure=Benjamini_Hochberg_procedure,
    Bonferroni_correction=Bonferroni_correction
)


def adjust_for_multiple_comparisons(p_values, alpha=0.05,
                                    method='Benjamini_Hochberg_procedure'):
    '''Corrects for multiple comparisons and returns the significant
    p-values.

    Parameters
    ----------
    p_values : array_like
    alpha : float, optional
        The expected proportion of false positive tests.
    method : string, optional
        Name of the method to use to correct for multiple comparisons.
        Options are "Benjamini_Hochberg_procedure", "Bonferroni_correction"
    Returns
    -------
    is_significant : boolean nd-array
        A boolean array the same shape as `p_values` indicating whether the
        null hypothesis has been rejected (True) or failed to reject
        (False).

    '''
    # TODO: add axis keyword?
    return MULTIPLE_COMPARISONS[method](p_values, alpha=alpha)


def fisher_z_transform(coherency1, bias1, coherency2=0, bias2=0):
    '''Transform the coherence magnitude to an approximately normal
    distribution.

    If two coherencies are provided, then the function returns the
    fisher z transform of the difference of the coherencies with
    `coherency1` - `coherency2`.

    Parameters
    ----------
    coherency1 : complex array
        The complex coherency between signals
    bias1 : float
        The bias from independent estimates of the frequency domain.
    coherency2 : complex array, optional
    bias2 : float, optional
        The bias from independent estimates of the frequency domain.

    Returns
    -------
    fisher_z_transform : real array
        Either the difference from 0 mean or, if another coherency is
        provided, the difference from that coherency.

    '''
    coherence_magnitude1 = np.abs(coherency1)
    coherence_magnitude1[coherence_magnitude1 >= 1] = (
        1 - np.finfo(float).eps)

    coherence_magnitude2 = np.array(np.abs(coherency2))
    coherence_magnitude2[coherence_magnitude2 >= 1] = (
        1 - np.finfo(float).eps)

    z1 = np.arctanh(coherence_magnitude1) - bias1
    z2 = np.arctanh(coherence_magnitude2) - bias2
    return (z1 - z2) / np.sqrt(bias1 + bias2)


def get_normal_distribution_p_values(data, mean=0, std_deviation=1):
    '''Given data, returns the probability the data was generated from
    a normal distribution with `mean` and `std_deviation`
    '''
    return 1 - norm.cdf(data, loc=mean, scale=std_deviation)


def coherence_bias(n_observations):
    ''' Enochson and Goodman (1965)
        Bokil et al. (2007)
    '''
    degrees_of_freedom = 2 * n_observations
    return 1.0 / (degrees_of_freedom - 2)
